from .results import BrokenCheckerException, BrokenServiceException, OfflineException
from requests import post
from os import environ
import json
from asyncio import TimeoutError
BACKEND = environ["CONNHANDLER_URL"]


def rpc_call(target, action_name, runlength, logger=None, **kwargs):
    try:
        if isinstance(action_name, type):
            action_name = type.__name__()
        
        kwargs["address"] = target
        kwargs.setdefault("initial_timeout", 10)
        kwargs.setdefault("long_timeout", runlength)

        # if logger is not None:
        #     logger.debug(kwargs)
        #     logger.debug(json.dumps(kwargs))
            
        req = post("{}/{}".format(BACKEND, action_name), json=kwargs)
        
        if logger is not None:
            logger.debug(f"RPC Result: {req.text}")

        result = req.json()

    except Exception as ex:
        if logger is not None:
            logger.error("Internal RPC Error", exc_info=ex)
        raise BrokenCheckerException

    if result['status'] == 'aborted':
        message = result['exception']['message']
        if logger is not None:
            logger.error("RPC {} {} did not return successfully {}".format(action_name, target, result))
        if result['exception']['type'] == BrokenServiceException.__name__:
            raise BrokenServiceException(message)
        if result['exception']['type'] == OfflineException.__name__:
            raise OfflineException(message)
        if 'Connection' in result['exception']['type']:
            raise ConnectionError(message)
        if result['exception']['type'] == TimeoutError.__name__:
            raise OfflineException(message)

    return result



from abc import ABC, abstractmethod

from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.background import BackgroundTask

from asyncio import wait_for, TimeoutError
from functools import partial
from traceback import format_tb

from sys import exc_info


class LR_Action(ABC):

    def __init__(self, request_params):
        self.request = request_params

    @abstractmethod
    async def initial_call(self):
        pass
    
    @abstractmethod
    async def background_call(self):
        pass

    @abstractmethod
    async def cleanup(self):
        pass


class LR_Handler(Starlette):
    def __init__(self, lr_callables, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for call in lr_callables:

            p_call = partial(self._callwrapper, call)
            self.add_route(f"/{call.__name__}", p_call, methods=["POST"])

            print(f"/{call.__name__}")

    @staticmethod
    async def _callwrapper(
         lr_callable, 
         scope, recieve, send, 
         *args, **rkwargs
         ):
        
        async def __bg_task(timeout):  # lr_action: LR_Action, 
            nonlocal lr_action
            try:
                await wait_for(lr_action.background_call(), timeout)
                print({
                      "status": "FINISHED",
                      "exception": None
                      })

            except TimeoutError:
                exc_inf = exc_info()
                print({
                      "status": "Timeout",
                      "exception": {
                        "type":    exc_inf[0].__name__,
                        "message": str(exc_inf[1]),
                        "trace ":  format_tb(exc_inf[2])
                        }
                      })

            except Exception:
                exc_inf = exc_info()
                print({
                      "status": "Error",
                      "exception": {
                        "type":    exc_inf[0].__name__,
                        "message": str(exc_inf[1]),
                        "trace ":  format_tb(exc_inf[2])
                        }
                      })

            finally:
                try:
                    await lr_action.cleanup()
                except Exception:
                    exc_inf = exc_info()
                    print({
                      "status": "cleanup failed",
                      "exception": {
                        "type":    exc_inf[0].__name__,
                        "message": str(exc_inf[1]),
                        "trace ":  format_tb(exc_inf[2])
                        }
                      })
                del lr_action

        request = Request(scope, recieve)
        kwargs = await request.json()

        lr_action = lr_callable(kwargs)
        # lr_action = None
        try:
            # initial Call
            continue_with_bg, ret_dict = await wait_for(lr_action.initial_call(), kwargs['initial_timeout'])
            print("initial call succeded")
            # Background Task
            
            if continue_with_bg:
                task = BackgroundTask(__bg_task, kwargs['long_timeout'])
            else:
                task = None

        except Exception:
            print("got Exception")
            cleanup_failed = None
            exc_inf = exc_info()
            try:
                await lr_action.cleanup()
            except Exception:
                cleanup_failed = exc_info()

            # print(await lr_action.reader.read(20000))
            if cleanup_failed is None:
                return await JSONResponse({
                    "status": "aborted",
                    "exception": {
                        "type":    exc_inf[0].__name__,
                        "message": str(exc_inf[1]),
                        "trace ":  format_tb(exc_inf[2])
                        },
                    "cleanup failed": None
                    })(scope, recieve, send)
            else:
                return await JSONResponse({
                    "status": "aborted",
                    "exception": {
                        "type":    exc_inf[0].__name__,
                        "message": str(exc_inf[1]),
                        "trace ":  format_tb(exc_inf[2])
                        },
                    "cleanup failed": {
                        "type":    cleanup_failed[0].__name__,
                        "message": str(cleanup_failed[1]),
                        "trace ":  format_tb(cleanup_failed[2])
                        }
                    })(scope, recieve, send)
        
        return await JSONResponse(
                        {"status": "OK", "result": ret_dict},
                        background=task
                        )(scope, recieve, send)





import collections
import configparser
import os
import logging

from functools import wraps
from pymongo.errors import PyMongoError
# import logging
from pymongo import MongoClient
from .results import BrokenCheckerException
from .utils import base64ify

# from urllib.parse import quote_plus

try:
    import uwsgi
    from uwsgidecorators import postfork
except (ImportError, ModuleNotFoundError):
    def postfork(func):
        func()
        return func

# LOGGING SETUP
logging.basicConfig(level=logging.DEBUG)
dictlogger = logging.Logger(__name__)
dictlogger.setLevel(logging.DEBUG)

# DIR
DB_DEFAULT_DIR = None
DB_GLOBAL_CACHE_SETTING = False
RETRY_COUNT = 4

# DB DEFAULT PARAMS
DB_DEFAULT_USER = 'root'
DB_DEFAULT_PASS = 'example'
DB_DEFAULT_HOST = '172.20.0.3'
DB_DEFAULT_PORT = 27017

# INIT OVERRIDE
print("READING INIT")
config = configparser.ConfigParser()
config.read("db.ini")
config.read("DB.ini")
config.read("database.ini")
config.read("Database.ini")
config.read("DATABASE.ini")

if 'DATABASE' in config:
    if 'HOST' in config['DATABASE']:
        DB_DEFAULT_HOST = config['DATABASE']['HOST']
    if 'PORT' in config['DATABASE']:
        DB_DEFAULT_PORT = int(config['DATABASE']['PORT'])
    if 'USER' in config['DATABASE']:
        DB_DEFAULT_USER = config['DATABASE']['USER']
    if 'PASSWORD' in config['DATABASE']:
        DB_DEFAULT_PASS = config['DATABASE']['PASSWORD']

if 'MONGO_HOST' in os.environ:
    DB_DEFAULT_HOST = os.environ['MONGO_HOST']
if 'MONGO_PORT' in os.environ:
    DB_DEFAULT_PORT = int(os.environ['MONGO_PORT'])
if 'MONGO_USER' in os.environ:
    DB_DEFAULT_USER = os.environ['MONGO_USER']
if 'MONGO_PASSWORD' in os.environ:
    DB_DEFAULT_PASS = os.environ['MONGO_PASSWORD']

print("host = ", DB_DEFAULT_HOST)
print("port = ", DB_DEFAULT_PORT)
print("username = ", DB_DEFAULT_USER)
print("password = ", DB_DEFAULT_PASS)

global CLIENT
CLIENT = MongoClient(
    host=DB_DEFAULT_HOST,
    port=DB_DEFAULT_PORT,
    username=DB_DEFAULT_USER,
    password=DB_DEFAULT_PASS)


@postfork
def initialize_connection():
    global CLIENT
    CLIENT = MongoClient(
        host=DB_DEFAULT_HOST,
        port=DB_DEFAULT_PORT,
        username=DB_DEFAULT_USER,
        password=DB_DEFAULT_PASS)
    print("MONGO CLIENT INITIALIZED")


def to_keyfmt(key):
    return str(key)  # + type(key).__name__


def _try_n_times(func):
    @wraps(func)
    def try_n_times(*args, **kwargs):
        for i in range(RETRY_COUNT):
            try:
                return func(*args, **kwargs)
            except PyMongoError as ex:
                dictlogger.error("noSQLdict_Error, Try {}".format(str(i)), exc_info=ex)
                if i == RETRY_COUNT:
                    raise

    return try_n_times


class StoredDict(collections.MutableMapping):
    """
    A dictionary that is MongoDb backed.
    """

    def __init__(self, checker_name="BaseChecker", dict_name="default",
                 host=DB_DEFAULT_HOST, port=DB_DEFAULT_PORT,
                 username=DB_DEFAULT_USER, password=DB_DEFAULT_PASS):
        global CLIENT
        for i in range(RETRY_COUNT):
            try:
                # self.client = 
                self.dict_name = base64ify(dict_name, altchars=b"-_")
                self.checker_name = checker_name
                #                   Table by checker
                self.db = CLIENT[checker_name][self.dict_name]
                #                           Collection by team/global
                self.cache = dict()

                # Add DB index
                try:
                    self.db.index_information()['checker_key']
                except KeyError:
                    self.db.create_index(
                        [("key", 1)],
                        name="checker_key", unique=True, background=True
                    )
            except PyMongoError as ex:
                dictlogger.error("noSQLdict_Error", exc_info=ex)
                if i == RETRY_COUNT - 1:
                    raise BrokenCheckerException from ex

    @_try_n_times
    def __setitem__(self, key, value):
        self.cache[key] = value

        query_dict = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name
        }

        to_insert = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name,
            "value": value
        }

        self.db.replace_one(query_dict, to_insert, upsert=True)

    @_try_n_times
    def __getitem__(self, key, print_result=False):

        if key in self.cache:
            return self.cache[key]

        to_extract = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name
        }

        result = self.db.find_one(to_extract)

        if print_result:
            print(result)

        if result:
            self.cache[key] = result['value']
            return result['value']
        raise KeyError()

    @_try_n_times
    def __delitem__(self, key):

        if key in self.cache:
            del self.cache[key]

        to_extract = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name
        }
        self.db.delete_one(to_extract)

    @_try_n_times
    def __len__(self):

        return self.db.count_documents(
            {
                "checker": self.checker_name,
                "name": self.dict_name}
        )

    @_try_n_times
    def __iter__(self):

        iterdict = {
            "checker": self.checker_name,
            "name": self.dict_name
        }
        results = self.db.find(iterdict)
        for key in map(lambda res: res['key'], results):
            yield key

    @_try_n_times
    def persist(self):

        self.cache = dict()

    @_try_n_times
    def __del__(self):
        self.persist()



from future.standard_library import install_aliases

install_aliases()

import atexit
import collections
import json
import os
import threading
import time
import logging
from functools import wraps
from typing import Any, Dict, Set, Iterator, Optional, TypeVar, cast

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path  # python 2 backport

from .utils import ensure_valid_filename, base64ify, debase64ify, start_daemon

logging.basicConfig(level=logging.DEBUG)
dictlogger = logging.Logger(__name__)
dictlogger.setLevel(logging.DEBUG)

T = TypeVar('T')  # typing things

DB_DEFAULT_DIR = os.path.join(os.getcwd(), ".data")  # Default location db files will be stored in -> usually cwd.
# TODO: Force remove locks after a while?
DB_LOCK_RETRYCOUNT = 6  # 2**6 / 10 seconds are 6.4 secs. -> That's how long the db will wait for a log
DB_PREFIX = "_store_"  # Prefix all db files will get
DB_EXTENSION = ".json"  # Extension all db files will get
DB_LOCK_EXTENSION = ".lock"  # Extension all lock folders will get
DB_GLOBAL_CACHE_SETTING = True

def makedirs(path, exist_ok=True):
    # type: (str, bool) -> None
    """
    Python2 ready
    param path: the path to create
    param exist_ok: ignore already existing path and do nothing
    """
    Path(path).mkdir(parents=True, exist_ok=exist_ok)


def _locked(func):
    # type: (T) -> T
    """
    Internal wrapper method for StoredDict that will ensure locks on a python threading level
    :param func: StoredDict method to be wrapped
    :return: the wrapped method
    """

    @wraps(func)
    def locked(self, *args, **kwargs):
        # type: (StoredDict, *str, **int)->Any
        """The called function first acquires a lock and then releases it later."""
        self.logger.debug("Locking {} db".format(self.name))
        self._local_lock.acquire()
        self.logger.debug("Log db lock for {}".format(self.name))
        try:
            return func(self, *args, **kwargs)
        finally:
            self._local_lock.release()
            self.logger.debug("Released db lock for {}".format(self.name))

    return cast(T, locked)


class StoredDict(collections.MutableMapping):
    """
    A dictionary that is filesystem backed.
    It will write to disk every few seconds and at exit.
    In case python crashes, changes may be gone. :/
    Note: Complex won't be tracked.
    """

    def __init__(self, base_path=DB_DEFAULT_DIR, name="default", persist_secs=3, ignore_locks=False, logger=None, *args,
                 **kwargs):
        # type: (str, str, int, bool, Optional[logging.Logger], *Any, **Any) -> None
        """
        Creates a new File System backed Store.
        It quacks like a dict and will persist to filesystem every few seconds if possible. :)
        :param base_path: the base path
        :param name: name of this store
        :param persist_secs: how often to persist dirty elements (0 means: never autostore. Call persist manually)
        :param ignore_locks: We usually write and read lock files before accessing the data.
                This flag seaves them out.
        :param logger: The logger instance to log events to
        """
        self._cache = {}  # type: Dict[str, any]
        self._dirties = set()  # type: Set[str]
        self._locks = set()  # type: Set[str]
        self._to_delete = set()  # type: Set[str]
        self._persist_thread = None  # type: Optional[threading.Thread]
        self.name = name
        self.path = os.path.join(base_path, ensure_valid_filename(name))  # type: str
        self.persist_secs = persist_secs  # type: int
        self.ignore_locks = ignore_locks  # type: bool
        self.persist_secs = persist_secs  # type: int

        if logger:
            self.logger = logger
        else:
            self.logger = dictlogger

        self._local_lock = threading.RLock()

        makedirs(self.path, exist_ok=True)
        atexit.register(self._cleanup)
        self._stopping = False

        self.update(dict(*args, **kwargs))  # In case we got initialized using a dict, make sure it's in sync.

    @_locked
    def _spawn_persist_thread(self):
        # type: () -> None
        """
        Spawns a thread persisting all changes, if necessary.
        """
        if self.persist_secs > 0 and not self._persist_thread:
            def persist_async():
                time.sleep(self.persist_secs)
                self.logger.debug("Persisting db {} from background.".format(self.name))
                self.persist()

            self._persist_thread = start_daemon(persist_async)

    @_locked
    def _cleanup(self):
        # type: () -> None
        """Cleans up the db: persists and releases all locks currently held."""
        self.logger.debug("StoredDict cleanup task running.")
        self._stopping = True
        self.persist()
        for lock in self._locks:
            self.release(lock)

    def __del__(self):
        # type: () -> None
        """
        Delete a key
        """
        self._cleanup()

    def _dir(self, key):
        # type: (str) -> str
        """The path for the this key"""
        return os.path.join(self.path, DB_PREFIX + base64ify(key, b"+-"))

    def _dir_jsonname(self, key):
        # type: (str) -> str
        """The path for the json db file for this key"""
        return "{}{}".format(self._dir(key), DB_EXTENSION)

    def _dir_lockname(self, key):
        # type: (str) -> str
        """The path for the lock file for this key"""
        return "{}{}".format(self._dir(key), DB_LOCK_EXTENSION)

    @_locked
    def release(self, locked_key):
        # type: (str) -> None
        """
        Release a file lock
        :param locked_key: the key we locked
        """
        if locked_key not in self._locks:
            raise KeyError("{} was not locked.".format(locked_key))
        self._dir_lockname(locked_key)
        os.rmdir(self._dir_lockname(locked_key))
        self._locks.remove(locked_key)

    @_locked
    def mark_dirty(self, key):
        # type: (str) -> Any
        """
        Manually mark an entry as dirty. It will be updated on disk on the next occasion
        :param key: the key that needs to be stored
        :return: the value contained in the key
        """
        val = self[key]
        self[key] = val
        return val

    def _create_lock_file(self, path, retrycount=DB_LOCK_RETRYCOUNT):
        # type: (str, int) -> None
        """Creates new lock file, waiting up to retrycount seconds. Raises TimeoutError if failed."""
        for i in range(0, retrycount):
            try:
                makedirs(path, exist_ok=False)
                return
            except OSError as ex:
                self.logger.debug("Waiting for lock on file {} (currently {})".format(path, ex))
                time.sleep(float(2 ** i) / 10)
        raise TimeoutError("Lock for {} could not be acquired in time!".format(path))

    @_locked
    def lock(self, key):
        # type: (str) -> None
        """
        Waits for a lock
        :param key: the key to lock
        """
        if key in self._locks:
            raise KeyError("{} already locked".format(key))
        self._create_lock_file(self._dir_lockname(key))
        self._locks.add(key)

    @_locked
    def is_locked(self, key):
        # type: (str) -> bool
        """
        Returns if the key is currently locked by this process
        :param key: The key
        :return: True if locked by this process, False otherwise
        """
        return key in self._locks

    @_locked
    def reload(self):
        # type: () -> None
        """
        Reloads stored values from disk.
        There is usually no reason to call this.
        Non persisted changes might be lost.
        Only reason would be if another process fiddles with our data concurrently.
        """
        self._cache = {}
        self._dirties = set()
        self._to_delete = set()

    @_locked
    def persist(self):
        # type: () -> None
        """
        Stores all dirty data to disk.
        If no data is to be stored, it's basically free to call.
        """
        for key in self._to_delete:
            locked = self.is_locked(key) or self.ignore_locks
            if not locked:
                self.lock(key)
            os.remove(self._dir_jsonname(key))
            if not locked:
                self.release(key)
        self._to_delete = set()

        for key in self._dirties:
            locked = self.is_locked(key)
            if not locked:
                self.lock(key)
            try:
                with open(self._dir_jsonname(key), "wb") as f:
                    f.write(json.dumps(self._cache[key]).encode("utf-8"))
            finally:
                if not locked:
                    self.release(key)
        self._dirties = set()

    @_locked
    def __getitem__(self, key):
        # type: (str) -> Any
        """
        Gets an item from the dict. Will hit the cache first, then disk.
        :param key: the key to look up
        :return: the value
        """
        if key in self._to_delete:
            raise KeyError("Key was marked for deletion: {}".format(key))
        try:
            return self._cache[key]
        except KeyError:
            self.logger.debug("Hitting disk for {}".format(key))

        locked = self.is_locked(key) or self.ignore_locks
        if not locked:
            self.lock(key)
        try:
            with open(self._dir_jsonname(key), "rb") as f:
                val = json.loads(f.read().decode("utf-8"))
        except IOError as ex:
            raise KeyError("Key {} not found - {}".format(key, ex))
        finally:
            if not locked:
                self.release(key)

        self._cache[key] = val
        return val

    @_locked
    def __setitem__(self, key, value):
        # type: (str, Any) -> None
        """
        Set an item. It'll be stored to disk on the next persist.
        :param key: Key to store
        :param value: Value to store
        """

        if key in self._to_delete:
            self._to_delete.remove(key)
        self._cache[key] = value
        self._dirties.add(key)

    @_locked
    def __delitem__(self, key):
        # type: (str) -> None
        """
        Delete an item. It will be deleted from disk on the next .persist().
        :param key: the key to delete
        """
        self._to_delete.add(key)

    def __iter__(self):
        # type: () -> Iterator[(str, Any)]
        """
        Iterates over the dict. Implicitly persisting the data before reading.
        :return: An iterator containing all keys to a dict.
        """
        self.persist()
        keys = [debase64ify(x[len(DB_PREFIX):-len(DB_EXTENSION)], b"+-") for x in os.listdir(self.path) if
                x.startswith(DB_PREFIX) and x.endswith(DB_EXTENSION)]
        for key in keys:
            yield key

    def __len__(self):
        # type: () -> int
        """
        Calculates the length. Implicitly calls persist.
        :return: the the number of elements
        """
        self.persist()
        keys = [x[len(DB_PREFIX):-len(DB_EXTENSION)] for x in os.listdir(self.path) if
                x.startswith(DB_PREFIX) and x.endswith(DB_EXTENSION)]
        return len(keys)




from abc import ABC
from enum import IntEnum

class Result(IntEnum):
    """
    Result Values to be returned from a Checker
    """
    INTERNAL_ERROR = -1  # The checker crashed
    OK = 0  # Everything is alright
    MUMBLE = 1  # (ENOFLAG/mumble)
    OFFLINE = 2  # It's dead, jim

    # noinspection PyTypeChecker
    @classmethod
    def is_valid(cls, value):
        # type: (int) -> bool
        """
        Returns if the value is part of this Enum
        :param value: the value
        :return: True, if value is part of this Enum
        """
        return any(value == item.value for item in cls)


class EnoException(Exception, ABC):
    """
    Base error including the Result. Raise a subclass of me once we know what to do.
    """
    result = Result.INTERNAL_ERROR  # type: int


class BrokenServiceException(EnoException):
    """
    Indicates a broken Service
    """
    result = Result.MUMBLE  # type: int


class OfflineException(EnoException):
    """
    Service was not reachable (at least once) during our checks
    """
    result = Result.OFFLINE  # type: int


class BrokenCheckerException(EnoException):
    """
    Shouldn't be raised ever since we catch all abstract Errors
    Used internally if something goes horribly wrong
    """
    result = Result.INTERNAL_ERROR  # type: int



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="enochecker",
    version="0.0.5",
    author="domenukk",
    author_email="dmaier@sect.tu-berlin.de",
    description="Library to build checker scripts for EnoEngine A/D CTF Framework in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ENOWARS/enochecker",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    install_requires=requirements,
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        # 'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',

    ],
    zip_safe=False,  # This might be needed for reqirements.txt
)



from future.standard_library import install_aliases

install_aliases()

from .utils import assert_equals, assert_in, \
    base64ify, debase64ify, sha256ify, \
    ensure_bytes, ensure_unicode, ensure_valid_filename, snake_caseify, \
    readline_expect, serve_once, start_daemon, SimpleSocket  # the util stuff
from .results import BrokenServiceException, EnoException, OfflineException, Result  # Possible results
from .enochecker import run, parse_args, BaseChecker  # the BaseChecker
from .checkerservice import CHECKER_METHODS

name = "enochecker"
